@@ -24,7 +24,6 @@ from flask import redirect
 from flask import render_template
 from flask import request, url_for
 from flask import abort
-from flask import Response
 from flask.views import MethodView
 from flask.json import JSONEncoder
 try:
@@ -33,11 +32,12 @@ except ImportError:
     RequestParser = None
 import jsonschema
 from mistune import markdown
-from .constants import OPTIONAL_FIELDS
+from .constants import OPTIONAL_FIELDS, OPTIONAL_OAS3_FIELDS
 from .utils import extract_definitions
 from .utils import get_specs
 from .utils import get_schema_specs
 from .utils import parse_definition_docstring
+from .utils import parse_imports
 from .utils import get_vendor_extension_fields
 from .utils import validate
 from .utils import LazyString
@@ -167,20 +167,50 @@ class Swagger(object):
     SCHEMA_LOCATIONS = {'query': 'args', 'header': 'headers',
                         'formData': 'form', 'body': 'json', 'path': 'path'}
 
+    def _init_config(self, config, merge):
+        """ Initializes self.config. If merge is set to true, then
+        self.config will be set to with config + DEFAULT_CONFIG.
+        """
+        if config and merge:
+            self.config = dict(self.DEFAULT_CONFIG.copy(), **config)
+        elif config and not merge:
+            self.config = config
+        elif not config:
+            self.config = self.DEFAULT_CONFIG.copy()
+        else:  # The above branches must be exhaustive
+            raise ValueError
+
     def __init__(
             self, app=None, config=None, sanitizer=None, template=None,
             template_file=None, decorators=None, validation_function=None,
-            validation_error_handler=None, parse=False):
+            validation_error_handler=None, parse=False, format_checker=None,
+            merge=False
+    ):
         self._configured = False
         self.endpoints = []
         self.definition_models = []  # not in app, so track here
         self.sanitizer = sanitizer or BR_SANITIZER
-        self.config = config or self.DEFAULT_CONFIG.copy()
+
+        self._init_config(config, merge)
+
         self.template = template
         self.template_file = template_file
         self.decorators = decorators
-        self.validation_function = validation_function
-        self.validation_error_handler = validation_error_handler
+        self.format_checker = format_checker or jsonschema.FormatChecker()
+
+        def default_validation_function(data, schema):
+            return jsonschema.validate(
+                data, schema, format_checker=self.format_checker,
+            )
+
+        def default_error_handler(e, _, __):
+            return abort(400, e.message)
+
+        self.validation_function = validation_function\
+            or default_validation_function
+
+        self.validation_error_handler = validation_error_handler\
+            or default_error_handler
         self.apispecs = {}  # cached apispecs
         self.parse = parse
         if app:
@@ -205,7 +235,6 @@ class Swagger(object):
                 raise RuntimeError('Please install flask_restful')
             self.parsers = {}
             self.schemas = {}
-            self.format_checker = jsonschema.FormatChecker()
             self.parse_request(app)
 
         self._configured = True
@@ -221,7 +250,8 @@ class Swagger(object):
         if filename.endswith('.json'):
             loader = json.load
         elif filename.endswith('.yml') or filename.endswith('.yaml'):
-            loader = yaml.safe_load
+            loader = lambda stream: \
+                yaml.safe_load(parse_imports(stream.read(), filename))
         else:
             with codecs.open(filename, 'r', 'utf-8') as f:
                 contents = f.read()
@@ -229,7 +259,8 @@ class Swagger(object):
                 if contents[0] in ['{', '[']:
                     loader = json.load
                 else:
-                    loader = yaml.safe_load
+                    loader = lambda stream: \
+                        yaml.safe_load(parse_imports(stream.read(), filename))
         with codecs.open(filename, 'r', 'utf-8') as f:
             return loader(f)
 
@@ -328,6 +359,21 @@ class Swagger(object):
             data["securityDefinitions"] = self.config.get(
                 'securityDefinitions'
             )
+
+        def is_openapi3():
+            """
+            Returns True if openapi_version is 3
+            """
+            return openapi_version and openapi_version.split('.')[0] == '3'
+
+        if is_openapi3():
+            # enable oas3 fields when openapi_version is 3.*.*
+            optional_oas3_fields = self.config.get(
+                'optional_oas3_fields') or OPTIONAL_OAS3_FIELDS
+            for key in optional_oas3_fields:
+                if self.config.get(key):
+                    data[key] = self.config.get(key)
+
         # set defaults from template
         if self.template is not None:
             data.update(self.template)
@@ -458,6 +504,22 @@ class Swagger(object):
                 except (KeyError, TypeError):
                     prefix = ''
                 srule = '{0}{1}'.format(prefix, rule)
+
+                try:
+                    # handle basePath
+                    base_path = self.template['basePath']
+
+                    if base_path:
+                        if base_path.endswith('/'):
+                            base_path = base_path[:-1]
+                        if base_path:
+                            # suppress base_path from srule if needed.
+                            # Otherwise we will get definitions twice...
+                            if srule.startswith(base_path):
+                                srule = srule[len(base_path):]
+                except (KeyError, TypeError):
+                    pass
+
                 # old regex '(<(.*?\:)?(.*?)>)'
                 for arg in re.findall('(<([^<>]*:)?([^<>]*)>)', srule):
                     srule = srule.replace(arg[0], '{%s}' % arg[2])
@@ -586,12 +648,14 @@ class Swagger(object):
                 schemas = self.schemas[path_key]
             else:
                 doc = None
+                definitions = None
                 for spec in self.config['specs']:
                     apispec = self.get_apispecs(endpoint=spec['endpoint'])
                     if path in apispec['paths']:
                         if request.method.lower() in apispec['paths'][path]:
                             doc = apispec['paths'][path][
                                 request.method.lower()]
+                            definitions = apispec.get('definitions', {})
                             break
                 if not doc:
                     return
@@ -604,13 +668,16 @@ class Swagger(object):
                     location = self.SCHEMA_LOCATIONS[param['in']]
                     if location == 'json':  # load data from 'request.json'
                         schemas[location] = param['schema']
+                        schemas[location]['definitions'] = dict(definitions)
                     else:
                         name = param['name']
                         if location != 'path':
                             parsers[location].add_argument(
                                 name,
                                 type=self.SCHEMA_TYPES[
-                                    param.get('type', None)],
+                                    param['schema'].get('type', None)
+                                    if 'schema' in param
+                                    else param.get('type', None)],
                                 required=param.get('required', False),
                                 location=self.SCHEMA_LOCATIONS[
                                     param['in']],
@@ -631,11 +698,9 @@ class Swagger(object):
                 parsed_data['json'] = request.json or {}
             for location, data in parsed_data.items():
                 try:
-                    jsonschema.validate(
-                        data, schemas[location],
-                        format_checker=self.format_checker)
+                    self.validation_function(data, schemas[location])
                 except jsonschema.ValidationError as e:
-                    abort(Response(e.message, status=400))
+                    self.validation_error_handler(e, data, schemas[location])
 
             setattr(request, 'parsed_data', parsed_data)
 
