@@ -19,6 +19,7 @@ from flask import abort
 from flask import current_app
 from flask import request
 from flask.views import MethodView
+
 from .constants import OPTIONAL_FIELDS
 from .marshmallow_apispec import SwaggerView
 from .marshmallow_apispec import convert_schemas
@@ -197,6 +198,12 @@ def swag_from(
     """
 
     def resolve_path(function, filepath):
+        try:
+            from pathlib import Path
+            if isinstance(filepath, Path):
+                filepath = str(filepath)
+        except ImportError:
+            pass
         if not filepath.startswith('/'):
             if not hasattr(function, 'root_path'):
                 function.root_path = get_root_path(function)
@@ -206,7 +213,7 @@ def swag_from(
 
     def set_from_filepath(function):
         final_filepath = resolve_path(function, specs)
-        function.swag_type = filetype or specs.split('.')[-1]
+        function.swag_type = filetype or final_filepath.split('.')[-1]
 
         if endpoint or methods:
             if not hasattr(function, 'swag_paths'):
@@ -227,9 +234,20 @@ def swag_from(
     def set_from_specs_dict(function):
         function.specs_dict = specs
 
+    def is_path(specs):
+        """ Returns True if specs is a string or pathlib.Path
+        """
+        is_str_path = isinstance(specs, string_types)
+        try:
+            from pathlib import Path
+            is_py3_path = isinstance(specs, Path)
+            return is_str_path or is_py3_path
+        except ImportError:
+            return is_str_path
+
     def decorator(function):
 
-        if isinstance(specs, string_types):
+        if is_path(specs):
             set_from_filepath(function)
             # function must have or a single swag_path or a list of them
             swag_path = getattr(function, 'swag_path', None)
@@ -258,19 +276,29 @@ def swag_from(
     return decorator
 
 
-def __replace_ref(schema, relative_path):
+def __replace_ref(schema, relative_path, swag):
     """ TODO: add dev docs
 
     :param schema:
     :param relative_path:
+    :param swag:
     :return:
     """
     absolute_path = os.path.dirname(sys.argv[0])
     new_value = {}
     for key, value in schema.items():
         if isinstance(value, dict):
-            new_value[key] = __replace_ref(value, relative_path)
+            new_value[key] = __replace_ref(value, relative_path, swag)
         elif key == '$ref':
+            # see:
+            # https://swagger.io/docs/specification/describing-request-body/
+            if len(value) > 2 and value.startswith('#/'):  # $ref is local
+                content = swag
+                for id in value.split('/')[1:]:
+                    content = content[id]
+                return __replace_ref(content, relative_path, swag) \
+                    if isinstance(content, dict) else content
+
             if len(value) > 0 and value[0] == '/':
                 file_ref_path = absolute_path + value
             else:
@@ -286,7 +314,7 @@ def __replace_ref(schema, relative_path):
                 content = yaml.safe_load((file_content[comment_index:]))
                 new_value = content
                 if isinstance(content, dict):
-                    new_value = __replace_ref(content, relative_path)
+                    new_value = __replace_ref(content, relative_path, swag)
         else:
             new_value[key] = value
     return new_value
@@ -407,7 +435,7 @@ def validate(
         relative_path = absolute_path
     else:
         relative_path = os.path.dirname(filepath)
-    main_def = __replace_ref(main_def, relative_path)
+    main_def = __replace_ref(main_def, relative_path, swag)
 
     try:
         validation_function(data, main_def)
